"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
from typing import Tuple
import _setup_test_env  # noqa
import numpy as np
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model


def test_pack_padded():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Dim]:
            pack, pack_dim = rf.pack_padded(x, dims=[batch_dim, time_dim], enforce_sorted=False)
            return pack, pack_dim

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, pack_dim = model(extern_data["data"])
        out.mark_as_default_output(shape=(pack_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_pack_padded_md():
    batch_dim = Dim(3, name="batch")
    hyps_dim = Dim(5, name="hyps")
    hyps_spatial_dim = Dim(Tensor("hyps_spatial", [batch_dim, hyps_dim], dtype="int32"))
    vocab_dim = Dim(1000, name="vocab")
    extern_data = TensorDict(
        {"labels": Tensor("labels", [batch_dim, hyps_dim, hyps_spatial_dim], dtype="int32", sparse_dim=vocab_dim)}
    )
    hyps_packed_spatial_dim = Dim(None, name="hyps_packed_spatial")

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        labels = extern_data["labels"]
        assert hyps_spatial_dim in labels.dims
        assert hyps_spatial_dim.dyn_size_ext.dims == (batch_dim, hyps_dim)
        labels.mark_as_output("labels", shape=(batch_dim, hyps_dim, hyps_spatial_dim))
        hyps_spatial_dim.dyn_size_ext.mark_as_output("hyps_spatial_size", shape=[batch_dim, hyps_dim])
        labels_, _ = rf.pack_padded(labels, dims=[hyps_dim, hyps_spatial_dim], out_dim=hyps_packed_spatial_dim)
        assert hyps_packed_spatial_dim.dyn_size_ext.dims == (batch_dim,)
        hyps_packed_spatial_dim.dyn_size_ext.mark_as_output("packed_size", shape=[batch_dim])
        labels_.mark_as_default_output(shape=[batch_dim, hyps_packed_spatial_dim])

    out = run_model(extern_data, lambda **_kwargs: rf.Module(), _forward_step, test_tensorflow=False)
    print("out:")
    print(out["output"].raw_tensor.shape)
    print(out["output"].raw_tensor)
    batch_size = out["labels"].raw_tensor.shape[0]
    for b in range(batch_size):
        num_hyps = hyps_dim.dimension
        hyps_spatial_sizes = out["hyps_spatial_size"].raw_tensor[b]  # [hyps]
        hyps_packed_spatial_size = out["packed_size"].raw_tensor[b]  # []
        print(f"batch {b}, hyps_spatial_size {hyps_spatial_sizes}, hyps_packed_spatial_size {hyps_packed_spatial_size}")
        assert hyps_packed_spatial_size == sum(hyps_spatial_sizes)
        labels = [out["labels"].raw_tensor[b, h, : hyps_spatial_sizes[h]] for h in range(num_hyps)]
        labels_ = [seq.tolist() for seq in labels]
        print("labels:", labels_)
        packed = out["output"].raw_tensor[b, :hyps_packed_spatial_size]  # [hyps_packed_spatial]
        packed_ = packed.tolist()
        print("packed:", packed_)
        assert sum(labels_, []) == packed_


def test_masked_select():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "mask": Tensor("mask", [batch_dim, time_dim], dtype="bool"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        x = extern_data["data"]
        mask = extern_data["mask"]
        out, pack_dim = rf.masked_select(x, mask=mask, dims=[batch_dim, time_dim])
        out.mark_as_default_output(shape=(pack_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_masked_select_broadcast():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    extern_data = TensorDict(
        {
            "data": Tensor("data", [time_dim], dtype="float32"),
            "mask": Tensor("mask", [batch_dim, time_dim], dtype="bool"),
        }
    )

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        x = extern_data["data"]  # [T]
        mask = rf.sequence_mask([batch_dim, time_dim], device=x.device)
        indices = rf.range_over_dim(time_dim, device=x.device)  # [T]->T
        out, num_states_dim = rf.masked_select(indices, mask=mask, dims=mask.dims)  # [T_]->T
        out.mark_as_default_output(shape=(num_states_dim,))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_masked_select_single_dim():
    # https://github.com/rwth-i6/returnn/issues/1605
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "mask": Tensor("mask", [batch_dim, time_dim], dtype="bool"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        x = extern_data["data"]
        mask = extern_data["mask"]
        out, pack_dim = rf.masked_select(x, mask=mask, dims=[time_dim])
        out.mark_as_default_output(shape=(batch_dim, pack_dim, in_dim))
        pack_dim.dyn_size_ext.mark_as_output("out_size", shape=[batch_dim])
        x.mark_as_output("input", shape=[batch_dim, time_dim, in_dim])
        mask.mark_as_output("mask", shape=[batch_dim, time_dim])
        time_dim.dyn_size_ext.mark_as_output("in_size", shape=[batch_dim])

    res = run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step, test_tensorflow=False)
    for k, v in res.data.items():
        print(f"{k} {v} = {v.raw_tensor}")
    num_batches = res["input"].raw_tensor.shape[0]
    have_non_zero = False
    have_less = False
    for b in range(num_batches):
        in_size = res["in_size"].raw_tensor[b]
        out_size = res["out_size"].raw_tensor[b]
        print(f"batch {b}, in_size {in_size}, out_size {out_size}")
        assert 0 <= out_size <= in_size
        if out_size > 0:
            have_non_zero = True
        if out_size < in_size:
            have_less = True
        c = 0
        for t in range(in_size):
            if res["mask"].raw_tensor[b, t]:
                c += 1
        assert c == out_size
        t_ = 0
        for t in range(in_size):
            if res["mask"].raw_tensor[b, t]:
                assert (res["input"].raw_tensor[b, t] == res["output"].raw_tensor[b, t_]).all()
                t_ += 1
    assert have_non_zero and have_less  # just that the test case covered all cases


def test_pad_packed():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        x = extern_data["data"]
        x.mark_as_output("in", shape=(batch_dim, time_dim, in_dim))
        x_, batch_time_flat_dim = rf.pack_padded(x, dims=[batch_dim, time_dim])
        assert x_.dims_set == {batch_time_flat_dim, in_dim}
        # test [batch_fime_flat,feat], batch_fime_flat -> [batch, time]
        x__ = rf.pad_packed(x_, in_dim=batch_time_flat_dim, dims=[batch_dim, time_dim])
        assert x__.dims_set == {batch_dim, time_dim, in_dim}
        x__.mark_as_output("out", shape=(batch_dim, time_dim, in_dim))

    out_dict = run_model(extern_data, lambda **_kwargs: rf.Module(), _forward_step, test_tensorflow=False)
    in_ = out_dict["in"]
    out = out_dict["out"]
    np.testing.assert_array_equal(in_.raw_tensor, out.raw_tensor)


def test_pad_packed_batched():
    batch_dim_ = Dim(3, name="batch")
    in_dim = Dim(2, name="in")

    # noinspection PyShadowingNames
    def _forward_step(**_kwargs):
        time1_dim = Dim(rf.convert_to_tensor(np.array([1, 3, 4]), name="time1", dims=[batch_dim_], dtype="int32"))
        time2_dim = Dim(rf.convert_to_tensor(np.array([3, 2, 2]), name="time2", dims=[batch_dim_], dtype="int32"))
        x = rf.random_uniform((batch_dim_, time1_dim, time2_dim, in_dim))
        x.mark_as_output("in", shape=(batch_dim_, time1_dim, time2_dim, in_dim))
        x_, time_flat_dim = rf.pack_padded(x, dims=[time1_dim, time2_dim])
        assert x_.dims_set == {batch_dim_, time_flat_dim, in_dim}
        x_.mark_as_output("flat", shape=(batch_dim_, time_flat_dim, in_dim))
        # test [batch,mult_fime_flat,feat], mult_fime_flat -> [time1, time2]
        x__ = rf.pad_packed(x_, in_dim=time_flat_dim, dims=[time1_dim, time2_dim])
        assert x__.dims_set == {batch_dim_, time1_dim, time2_dim, in_dim}
        x__.mark_as_output("out", shape=(batch_dim_, time1_dim, time2_dim, in_dim))

    out_dict = run_model(TensorDict(), lambda **_kwargs: rf.Module(), _forward_step, test_tensorflow=False)
    in_ = out_dict["in"]
    flat = out_dict["flat"]
    out = out_dict["out"]
    print("in:", in_, in_.raw_tensor.shape)
    print("in time1:", in_.dims[1].dyn_size)
    print("in time2:", in_.dims[2].dyn_size)
    print("flat:", flat, flat.raw_tensor.shape)
    print("out:", out, out.raw_tensor.shape)
    print("in raw:")
    print(in_.raw_tensor)
    print("flat raw:")
    print(flat.raw_tensor)
    print("out raw:")
    print(out.raw_tensor)
    np.testing.assert_array_equal(in_.raw_tensor, out.raw_tensor)


def test_masked_select_masked_scatter_vs_where_rev_dims():
    """
    Compare rf.where vs rf.masked_select+rf.masked_scatter.
    Some op (e.g. LM update) could be done more efficiently on just the packed data (rf.masked_select),
    that is why rf.masked_select+rf.masked_scatter can be useful over just using rf.where.
    (In general, when computing the new ``a`` is expensive.)
    The test does not cover this part on the computation (we just feed in some random ``a``),
    but it checks that the results are the same,
    as we had some problems with that in the past.
    """
    # noinspection PyShadowingNames
    batch_dim = Dim(2, name="batch")
    beam_dim = Dim(3, name="beam")

    extern_data = TensorDict(
        {
            "mask": Tensor("mask", [beam_dim, batch_dim], dtype="bool"),
            # Note: The dim order is relevant for this test. It passes when it is [batch_dim, beam_dim]...
            "a": Tensor("a", [beam_dim, batch_dim], dtype="int32"),
            "b": Tensor("b", [beam_dim, batch_dim], dtype="int32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        mask = extern_data["mask"]
        a = extern_data["a"]
        b = extern_data["b"]
        a.mark_as_output("a", shape=[beam_dim, batch_dim])
        b.mark_as_output("b", shape=[beam_dim, batch_dim])
        mask.mark_as_output("mask", shape=[beam_dim, batch_dim])

        # Code via rf.where.
        res_where = rf.where(mask, a, b)
        assert res_where.dims_set == {beam_dim, batch_dim}
        res_where.mark_as_output("res_where", shape=[beam_dim, batch_dim])

        # Code via rf.masked_select and rf.masked_scatter.
        a_packed, packed_dim = rf.masked_select(a, mask=mask, dims=[batch_dim, beam_dim])
        assert a_packed.dims_set == {packed_dim}
        res_packed = rf.masked_scatter(a_packed, b, mask=mask, dims=[batch_dim, beam_dim], in_dim=packed_dim)
        assert res_packed.dims_set == {batch_dim, beam_dim}
        res_packed.mark_as_output("res_packed", shape=[beam_dim, batch_dim])

    out_dict = run_model(extern_data, lambda **_kwargs: rf.Module(), _forward_step, test_tensorflow=False)
    res_where = out_dict["res_where"]
    res_packed = out_dict["res_packed"]
    assert res_where.raw_tensor.shape == res_packed.raw_tensor.shape
    print("a:")
    print(out_dict["a"].raw_tensor)
    print("b:")
    print(out_dict["b"].raw_tensor)
    print("mask:")
    print(out_dict["mask"].raw_tensor)
    print("result with where:")
    print(res_where.raw_tensor)
    print("result with packing:")
    print(res_packed.raw_tensor)
    np.testing.assert_equal(res_where.raw_tensor, res_packed.raw_tensor)


def test_masked_select_masked_scatter_vs_where_md_rev_dims():
    """
    Like :func:`test_masked_select_masked_scatter_vs_where_rev_dims`
    but we add another spatial dim, which then needs some further handling.
    """
    # noinspection PyShadowingNames
    batch_dim = Dim(2, name="batch")
    beam_dim = Dim(3, name="beam")
    hist_a_dim = Dim(Tensor("hist_a", [batch_dim, beam_dim], dtype="int32"))
    hist_b_dim = Dim(Tensor("hist_b", [batch_dim, beam_dim], dtype="int32"))

    extern_data = TensorDict(
        {
            "mask": Tensor("mask", [beam_dim, batch_dim], dtype="bool"),
            "a": Tensor("a", [beam_dim, batch_dim, hist_a_dim], dtype="int32"),
            "b": Tensor("b", [beam_dim, batch_dim, hist_b_dim], dtype="int32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        mask = extern_data["mask"]
        a = extern_data["a"]
        b = extern_data["b"]
        a.mark_as_output("a", shape=[beam_dim, batch_dim, hist_a_dim])
        b.mark_as_output("b", shape=[beam_dim, batch_dim, hist_b_dim])
        mask.mark_as_output("mask", shape=[beam_dim, batch_dim])
        hist_a_size = hist_a_dim.get_size_tensor()
        hist_b_size = hist_b_dim.get_size_tensor()

        # Code via rf.where.
        hist_comb_sizes = rf.where(mask, hist_a_size, hist_b_size)
        hist_comb_dim = Dim(hist_comb_sizes, name="hist_comb")
        a_ = rf.replace_dim_v2(a, in_dim=hist_a_dim, out_dim=hist_comb_dim)
        b_ = rf.replace_dim_v2(b, in_dim=hist_b_dim, out_dim=hist_comb_dim)
        res_where = rf.where(mask, a_, b_)
        assert res_where.dims_set == {beam_dim, batch_dim, hist_comb_dim}
        res_where.mark_as_output("res_where", shape=[beam_dim, batch_dim, hist_comb_dim])

        # Code via rf.masked_select and rf.masked_scatter.
        hist_a_packed_size, packed_dim = rf.masked_select(hist_a_size, mask=mask, dims=[batch_dim, beam_dim])
        hist_a_packed_dim = Dim(hist_a_packed_size, name="hist_a_packed")
        a_packed, _ = rf.masked_select(a, mask=mask, dims=[batch_dim, beam_dim], out_dim=packed_dim)
        a_packed = rf.replace_dim_v2(a_packed, in_dim=hist_a_dim, out_dim=hist_a_packed_dim)
        assert a_packed.dims_set == {packed_dim, hist_a_packed_dim}
        hist_merged_size = rf.masked_scatter(
            hist_a_packed_size, hist_b_size, mask=mask, dims=[batch_dim, beam_dim], in_dim=packed_dim
        )
        hist_merged_dim = Dim(hist_merged_size, name="hist_merged")
        a_packed = rf.replace_dim_v2(a_packed, in_dim=hist_a_packed_dim, out_dim=hist_merged_dim)
        b_packed = rf.replace_dim_v2(b, in_dim=hist_b_dim, out_dim=hist_merged_dim)
        res_packed = rf.masked_scatter(a_packed, b_packed, mask=mask, dims=[batch_dim, beam_dim], in_dim=packed_dim)
        assert res_packed.dims_set == {batch_dim, beam_dim, hist_merged_dim}
        res_packed.mark_as_output("res_packed", shape=[beam_dim, batch_dim, hist_merged_dim])

    out_dict = run_model(extern_data, lambda **_kwargs: rf.Module(), _forward_step, test_tensorflow=False)
    res_where = out_dict["res_where"]
    res_packed = out_dict["res_packed"]
    hist_where_dim = res_where.dims[-1]
    hist_packed_dim = res_packed.dims[-1]
    assert hist_where_dim.dyn_size_ext.dims_set == hist_packed_dim.dyn_size_ext.dims_set == {batch_dim, beam_dim}
    hist_where_size_raw = hist_where_dim.dyn_size_ext.copy_compatible_to_dims_raw((batch_dim, beam_dim))
    hist_packed_size_raw = hist_packed_dim.dyn_size_ext.copy_compatible_to_dims_raw((batch_dim, beam_dim))
    assert (hist_where_size_raw == hist_packed_size_raw).all()
    assert res_where.raw_tensor.shape == res_packed.raw_tensor.shape
    print("a:")
    print(out_dict["a"].raw_tensor)
    print("b:")
    print(out_dict["b"].raw_tensor)
    print("mask:")
    print(out_dict["mask"].raw_tensor)
    print("result with where:")
    print(res_where.raw_tensor)
    print("result with packing:")
    print(res_packed.raw_tensor)
    np.testing.assert_equal(res_where.raw_tensor, res_packed.raw_tensor)


def test_reshape():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.reshape(x, in_dims=(time_dim, in_dim), out_dims=(in_dim, time_dim))

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    # Note: The tested op here is a bit meaningless. It also is not consinstent for different batch sizes...
    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step, test_single_batch_entry=False)


def test_expand_dim():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    expand_dim = Dim(3, name="expand")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.expand_dim(x, expand_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, expand_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_concat():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Dim]:
            return rf.concat((x, in_dim), (x, in_dim))

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, dim = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_concat_partly_dyn_dim():
    time_static_dim = Dim(5, name="time_static")
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "left": Tensor("left", [batch_dim, time_static_dim, in_dim], dtype="float32"),
            "right": Tensor("right", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        left, right = extern_data["left"], extern_data["right"]
        out, out_time_dim = rf.concat((left, time_static_dim), (right, time_dim))
        out.mark_as_default_output(shape=(batch_dim, out_time_dim, in_dim))

    run_model(extern_data, lambda **_: rf.Module(), _forward_step)


def test_concat_dyn_time():
    time1_dim = Dim(Tensor("time1", [batch_dim], dtype="int32"))
    time2_dim = Dim(Tensor("time2", [batch_dim], dtype="int32"))
    extern_data = TensorDict(
        {
            "left": Tensor("left", [batch_dim, time1_dim], dtype="float32"),
            "right": Tensor("right", [batch_dim, time2_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        left, right = extern_data["left"], extern_data["right"]
        out, out_time_dim = rf.concat((left, time1_dim), (right, time2_dim))
        out.mark_as_default_output(shape=(batch_dim, out_time_dim))

    # test_single_batch_entry should test the interesting case.
    run_model(extern_data, lambda **_: rf.Module(), _forward_step, test_tensorflow=False)


def test_pad():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Tuple[Dim, Dim]]:
            pack, (new_time, new_feat) = rf.pad(x, axes=[time_dim, in_dim], padding=[(1, 2), (3, 4)], value=0)
            return pack, (new_time, new_feat)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, (new_time, new_feat) = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, new_time, new_feat))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_pad_time():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Tuple[Dim]]:
            pack, (new_time,) = rf.pad(x, axes=[time_dim], padding=[(1, 0)], value=0)
            return pack, (new_time,)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, (new_time,) = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, new_time, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_pad_time_right():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Tuple[Dim]]:
            pack, (new_time,) = rf.pad(x, axes=[time_dim], padding=[(0, 1)], value=1)
            return pack, (new_time,)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        data = extern_data["data"]
        data.mark_as_output("data", shape=(batch_dim, time_dim, in_dim))
        out, (new_time,) = model(data)
        out.mark_as_default_output(shape=(batch_dim, new_time, in_dim))

    res = run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)
    data_: Tensor = res["data"]
    out_: Tensor = res["output"]
    assert data_.dims == (batch_dim, time_dim, in_dim)
    new_time_dim = out_.dims[1]
    assert out_.dims == (batch_dim, new_time_dim, in_dim) and new_time_dim != time_dim
    # assert new_time_dim == time_dim + 1  # math dim... not really necessary check here...
    assert time_dim.dyn_size_ext.dims == new_time_dim.dyn_size_ext.dims == (batch_dim,)
    batch_size = batch_dim.get_dim_value()
    assert batch_size > 1
    assert len(set(time_dim.dyn_size_ext.raw_tensor)) > 1  # not all the same
    for b in range(batch_size):
        seq_len = time_dim.dyn_size_ext.raw_tensor[b]
        new_seq_len = new_time_dim.dyn_size_ext.raw_tensor[b]
        print(f"batch {b}, seq_len {seq_len}, new_seq_len {new_seq_len}")
        assert new_seq_len == seq_len + 1
        np.testing.assert_allclose(data_.raw_tensor[b, :seq_len], out_.raw_tensor[b, :seq_len])
        print(out_.raw_tensor[b])
        assert all(out_.raw_tensor[b, seq_len] == 1.0)


def test_pad_time_right_non_scalar():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "value": Tensor("value", [batch_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        data, value = extern_data["data"], extern_data["value"]
        data.mark_as_output("data", shape=(batch_dim, time_dim, in_dim))
        value.mark_as_output("value", shape=(batch_dim,))
        out, (new_time,) = rf.pad(data, axes=[time_dim], padding=[(0, 1)], value=value)
        out.mark_as_default_output(shape=(batch_dim, new_time, in_dim))

    # TF-layers currently does not support this.
    res = run_model(extern_data, lambda **_kwargs: rf.Module(), _forward_step, test_tensorflow=False)
    data_: Tensor = res["data"]
    value_: Tensor = res["value"]
    out_: Tensor = res["output"]
    assert data_.dims == (batch_dim, time_dim, in_dim)
    new_time_dim = out_.dims[1]
    assert out_.dims == (batch_dim, new_time_dim, in_dim) and new_time_dim != time_dim
    assert time_dim.dyn_size_ext.dims == new_time_dim.dyn_size_ext.dims == (batch_dim,)
    batch_size = batch_dim.get_dim_value()
    assert batch_size > 1
    assert len(set(time_dim.dyn_size_ext.raw_tensor)) > 1  # not all the same
    for b in range(batch_size):
        seq_len = time_dim.dyn_size_ext.raw_tensor[b]
        new_seq_len = new_time_dim.dyn_size_ext.raw_tensor[b]
        print(f"batch {b}, seq_len {seq_len}, new_seq_len {new_seq_len}")
        assert new_seq_len == seq_len + 1
        np.testing.assert_allclose(data_.raw_tensor[b, :seq_len], out_.raw_tensor[b, :seq_len])
        print(out_.raw_tensor[b])
        assert all(out_.raw_tensor[b, seq_len] == value_.raw_tensor[b])


def test_stack():
    batch_dim_ = Dim(3, name="batch")
    time_dim = Dim(5, name="time")

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        seq = rf.range_over_dim(time_dim)  # [T]
        out, _ = rf.stack([seq, seq, seq], out_dim=batch_dim_)  # [B,T]
        out.mark_as_default_output(shape=(batch_dim_, time_dim))

    out = run_model(TensorDict(), lambda *, epoch, step: rf.Module(), _forward_step, test_tensorflow=False)
    out = out["output"]
    assert out.dims == (batch_dim_, time_dim)
    assert out.raw_tensor.tolist() == [[0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 1, 2, 3, 4]]


def test_gather():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.gather(x, indices=0, axis=time_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_gather_2d_indices():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "classes": Tensor("classes", [batch_dim, time_dim], dtype="int32", sparse_dim=in_dim),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor, y: Tensor) -> Tensor:
            return rf.gather(x, indices=y, axis=in_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"], extern_data["classes"])
        out.mark_as_default_output(shape=(batch_dim, time_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_gather_feature_dim():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], feature_dim=in_dim, dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.gather(x, indices=0, axis=time_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        assert out.feature_dim == in_dim
        out.mark_as_default_output(shape=(batch_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_gather_time_static_clip_to_valid():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data_template = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], feature_dim=in_dim, dtype="float32"),
        }
    )

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        x = extern_data["data"]
        out = rf.gather(x, indices=0, axis=time_dim, clip_to_valid=True)
        out.mark_as_default_output(shape=(batch_dim, in_dim))

    run_model(extern_data_template, lambda *, epoch, step: rf.Module(), _forward_step)


def test_gather_3d_embed():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    num_embeddings_dim = Dim(2, name="num_embeddings")
    embed_dim = Dim(11, name="embed")
    extern_data_template = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], sparse_dim=in_dim, dtype="int32"),
            "embed": Tensor("embed", [in_dim, num_embeddings_dim, embed_dim], dtype="float32"),
        }
    )

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        x, embed = extern_data["data"], extern_data["embed"]
        out = rf.gather(embed, indices=x)
        out.mark_as_default_output(shape=(batch_dim, time_dim, num_embeddings_dim, embed_dim))
        rf.reduce_sum(out, axis=out.dims).mark_as_output("loss")

    run_model(extern_data_template, lambda *, epoch, step: rf.Module(), _forward_step)


def test_scatter_fill_inf():
    batch_dim_ = Dim(3, name="batch")
    states_dim = Dim(7, name="states")

    def _forward_step(**_kwargs):
        start_states = rf.convert_to_tensor(
            [2, 4, 5], name="start_states", dims=[batch_dim_], sparse_dim=states_dim, dtype="int32"
        )
        batch_dim_.get_size_tensor().mark_as_output("batch_size", shape=[])
        start_states.mark_as_output("start_states", shape=[batch_dim_])
        scores = rf.scatter(
            rf.zeros([batch_dim_]),
            indices=start_states,
            indices_dim=[batch_dim_],
            fill_value=float("-inf"),
        )  # [S], per state
        scores.mark_as_default_output(shape=[states_dim])

    res = run_model(
        TensorDict(),
        lambda *, epoch, step: rf.Module(),
        _forward_step,
        test_tensorflow=False,
        allow_inf_nan_in_output=True,
    )
    batch_size = res["batch_size"].raw_tensor.item()
    assert res["start_states"].raw_tensor.shape == (batch_size,)
    assert res["output"].raw_tensor.shape == (states_dim.dimension,)
    assert res["output"].raw_tensor.tolist().count(0.0) == batch_size
    assert res["output"].raw_tensor.tolist().count(float("-inf")) == states_dim.dimension - batch_size
    assert states_dim.dimension > batch_size
    for i in range(states_dim.dimension):
        if i in res["start_states"].raw_tensor:
            assert res["output"].raw_tensor[i] == 0.0
        else:
            assert res["output"].raw_tensor[i] == float("-inf")


def test_slice():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Dim]:
            pack, new_time = rf.slice(x, axis=time_dim, start=1, size=2)
            return pack, new_time

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, new_time = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, new_time, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_slice_dyn_size():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        x = extern_data["data"]
        size = time_dim.get_size_tensor()
        size = rf.reduce_min(size, axis=size.dims)  # do whatever, but get some scalar
        pack, new_time = rf.slice(x, axis=time_dim, size=size)
        pack.mark_as_default_output(shape=(batch_dim, new_time, in_dim))

    run_model(extern_data, lambda **_kwargs: rf.Module(), _forward_step)


def test_slice_batched_start_size():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "start_factor": Tensor("start_factor", [batch_dim], dtype="float32"),
            "size_factor": Tensor("size_factor", [batch_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        x = extern_data["data"]
        x.mark_as_output("data", shape=(batch_dim, time_dim, in_dim))
        start_factor = rf.sigmoid(extern_data["start_factor"])
        size_factor = rf.sigmoid(extern_data["size_factor"])
        start = start_factor * rf.cast(rf.clip_by_value(time_dim.get_size_tensor() - 2, 0), "float32")
        start = rf.cast(start, dtype="int32")
        start.mark_as_output("start", shape=(batch_dim,))
        size = size_factor * rf.cast(time_dim.get_size_tensor() - start, "float32")
        size = rf.cast(size, dtype="int32")
        size.mark_as_output("size", shape=(batch_dim,))
        pack, new_time = rf.slice(x, axis=time_dim, start=start, size=size)
        pack.mark_as_default_output(shape=(batch_dim, new_time, in_dim))

    res = run_model(extern_data, lambda **_kwargs: rf.Module(), _forward_step)
    data_: Tensor = res["data"]
    start_: Tensor = res["start"]
    size_: Tensor = res["size"]
    out_: Tensor = res["output"]
    assert data_.dims == (batch_dim, time_dim, in_dim)
    new_time_dim = out_.dims[1]
    assert out_.dims == (batch_dim, new_time_dim, in_dim) and new_time_dim != time_dim
    assert time_dim.dyn_size_ext.dims == new_time_dim.dyn_size_ext.dims == (batch_dim,)
    batch_size = batch_dim.get_dim_value()
    for b in range(batch_size):
        seq_len = time_dim.dyn_size_ext.raw_tensor[b]
        start = start_.raw_tensor[b]
        size = size_.raw_tensor[b]
        new_seq_len = new_time_dim.dyn_size_ext.raw_tensor[b]
        print(f"batch {b}, seq_len {seq_len}, start {start}, size {size}, new_seq_len {new_seq_len}")
        assert new_seq_len == size
        np.testing.assert_array_equal(data_.raw_tensor[b, start : start + size], out_.raw_tensor[b, :new_seq_len])


def test_slice_neg_end():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        x = extern_data["data"]
        pack, new_time = rf.slice(x, axis=time_dim, end=-1)
        pack.mark_as_default_output(shape=(batch_dim, new_time))

    run_model(extern_data, lambda **_kwargs: rf.Module(), _forward_step)
    run_model(extern_data, lambda **_kwargs: rf.Module(), _forward_step, dyn_dim_max_sizes={time_dim: 1})


def test_slice_start_zero_tensor():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], dtype="int32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        x = extern_data["data"]
        start = rf.zeros([batch_dim], dtype="int32")
        y, time_dim_ = rf.slice(x, axis=time_dim, start=start)
        assert time_dim_ != time_dim  # diff dim tag, even though the size is expected to be the same
        x.mark_as_output("data", shape=[batch_dim, time_dim])
        y.mark_as_default_output(shape=[batch_dim, time_dim_])

    res = run_model(extern_data, lambda **_kwargs: rf.Module(), _forward_step)
    data_: Tensor = res["data"]
    out_: Tensor = res["output"]
    np.testing.assert_array_equal(data_.raw_tensor, out_.raw_tensor)


def test_shift_right():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], sparse_dim=in_dim, dtype="int32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.shift_right(x, axis=time_dim, pad_value=0)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_reverse_sequence():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.reverse_sequence(x, axis=time_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_reverse_sequence_no_dyn():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        out = rf.reverse_sequence(extern_data["data"], axis=time_dim, handle_dynamic_dims=False)
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step, test_single_batch_entry=False)


def test_where():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "cond": Tensor("cond", [batch_dim, time_dim], dtype="bool"),
            "true": Tensor("true", [batch_dim, time_dim, in_dim], dtype="float32"),
            "false": Tensor("false", [batch_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        out = rf.where(extern_data["cond"], extern_data["true"], extern_data["false"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_search_sorted():
    batch_dim_ = Dim(3, name="batch")
    time_dim = Dim(13, name="time")

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        sorted_seq = rf.range_over_dim(time_dim, dtype="float32")  # [T]
        index1 = rf.search_sorted(sorted_seq, rf.constant(4.5, dims=()), axis=time_dim)  # [] -> T
        assert index1.dims == () and index1.sparse_dim == time_dim and index1.dtype == "int32"
        index1.mark_as_output("index1", shape=())
        index2 = rf.search_sorted(sorted_seq, rf.constant(4.0, dims=()), axis=time_dim)  # [] -> T
        assert index2.dims == () and index2.sparse_dim == time_dim and index2.dtype == "int32"
        index2.mark_as_output("index2", shape=())
        index3 = rf.search_sorted(sorted_seq, rf.constant(4.0, dims=()), axis=time_dim, side="right")  # [] -> T
        assert index3.dims == () and index3.sparse_dim == time_dim and index3.dtype == "int32"
        index3.mark_as_output("index3", shape=())
        index4 = rf.search_sorted(
            sorted_seq, rf.convert_to_tensor(np.array([4.0, 4.5, 5.0]), dims=[batch_dim_]), axis=time_dim
        )  # [] -> T
        assert index4.dims == (batch_dim_,) and index4.sparse_dim == time_dim and index4.dtype == "int32"
        index4.mark_as_output("index4", shape=(batch_dim_,))

    out = run_model(TensorDict(), lambda *, epoch, step: rf.Module(), _forward_step, test_tensorflow=False)
    assert out["index1"].dims == () and out["index1"].raw_tensor.item() == 5
    assert out["index2"].dims == () and out["index2"].raw_tensor.item() == 4
    assert out["index3"].dims == () and out["index3"].raw_tensor.item() == 5
    assert out["index4"].dims == (batch_dim_,) and out["index4"].raw_tensor.tolist() == [4, 5, 5]


def test_where_int():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "cond": Tensor("cond", [batch_dim, time_dim], dtype="bool"),
            "true": Tensor("true", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        out = rf.where(extern_data["cond"], extern_data["true"], 0)
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_copy_masked():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Conv1d, extern_data: TensorDict):
        x = extern_data["data"]
        x = x.copy_masked(1)
        # Do some pooling to make sure the copy_masked has an effect on the output.
        x, _ = rf.pool1d(x, mode="avg", pool_size=3, strides=1, padding="same", in_spatial_dim=time_dim)
        x.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step, test_single_batch_entry=False)


def test_cast_sparse():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Conv1d, extern_data: TensorDict):
        x = rf.reduce_argmax(extern_data["data"], axis=in_dim)
        assert x.sparse_dim == in_dim
        x.mark_as_output("argmax", shape=[batch_dim, time_dim])
        rf.cast(x, "float32").mark_as_output("float", shape=[batch_dim, time_dim])

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_repeat():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(3, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "repeats": Tensor("repeats", [batch_dim, time_dim], dtype="int32"),
        }
    )

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, extern_data: TensorDict, **_other):
        x = extern_data["data"]
        repeats = rf.abs(extern_data["repeats"]) % 4
        x.mark_as_output("data", shape=(batch_dim, time_dim, in_dim))
        repeats.mark_as_output("repeats", shape=(batch_dim, time_dim))
        out, out_spatial_dim = rf.repeat(x, repeats=repeats, in_spatial_dim=time_dim)
        out.mark_as_default_output(shape=(batch_dim, out_spatial_dim, in_dim))

    run_model(extern_data, lambda **_any: rf.Module(), _forward_step, test_tensorflow=False)

    # Again but with 2D data.
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], dtype="int32"),
            "repeats": Tensor("repeats", [batch_dim, time_dim], dtype="int32"),
        }
    )

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, extern_data: TensorDict, **_other):
        x = extern_data["data"]
        repeats = rf.abs(extern_data["repeats"]) % 4
        x.mark_as_output("data", shape=(batch_dim, time_dim))
        repeats.mark_as_output("repeats", shape=(batch_dim, time_dim))
        out, out_spatial_dim = rf.repeat(x, repeats=repeats, in_spatial_dim=time_dim)
        out.mark_as_default_output(shape=(batch_dim, out_spatial_dim))

    # Iterate over a few random seeds.
    # Make sure we covered some relevant cases.
    have_intermediate_zero = False
    have_leading_zero = False
    have_trailing_zero = False
    for seed in range(42, 45):
        res = run_model(
            extern_data, lambda **_any: rf.Module(), _forward_step, data_rnd_seed=seed, test_tensorflow=False
        )
        num_batch = res["data"].raw_tensor.shape[0]
        print(f"Seed {seed}, num_batch {num_batch}")
        for b in range(num_batch):
            in_seq_len = res["data"].dims[1].dyn_size_ext.raw_tensor[b]
            out_seq_len = res["output"].dims[1].dyn_size_ext.raw_tensor[b]
            data_b = res["data"].raw_tensor[b, :in_seq_len]
            repeats_b = res["repeats"].raw_tensor[b, :in_seq_len]
            if len(repeats_b) > 0 and repeats_b[0] == 0:
                have_leading_zero = True
            if len(repeats_b) > 0 and repeats_b[-1] == 0:
                have_trailing_zero = True
            if len(repeats_b) > 2 and any(
                repeats_b[i] == 0 and repeats_b[i - 1] != 0 and repeats_b[i + 1] != 0
                for i in range(1, len(repeats_b) - 2)
            ):
                have_intermediate_zero = True
            out_b = res["output"].raw_tensor[b, :out_seq_len]
            print(f" Batch {b}: data {data_b}, repeats {repeats_b}, out {out_b}")
            out_idx = 0
            for t in range(data_b.shape[0]):
                repeat_count = repeats_b[t]
                for _ in range(repeat_count):
                    assert out_b[out_idx] == data_b[t]
                    out_idx += 1
            assert out_idx == out_b.shape[0]
    assert have_intermediate_zero and have_leading_zero and have_trailing_zero



"""
Customized (derived) dict to pass as ``collected_outputs`` to some of the RF modules,
or potential other use cases.

You can predefine (by pattern) what kind of outputs you want to collect and store in this dict.
"""

from typing import Optional, Union, Sequence
import fnmatch


class CollectOutputsDict(dict):
    """
    Customized (derived) dict, where you can predefine (by key pattern)
    what kind of keys you want to collect and store in this dict.
    Other keys will be ignored.
    """

    def __init__(self, *args, allowed_key_patterns: Optional[Sequence[str]] = None, **kwargs):
        """
        Initialize the CollectOutputsDict.

        :param allowed_key_patterns:
            List of key patterns (with wildcards) that are allowed to be stored in the dict.
            If None, all keys are allowed.
        """
        super().__init__(*args, **kwargs)
        self.allowed_key_patterns = allowed_key_patterns

    def __setitem__(self, key, value):
        """
        Set an item in the dict if the key matches allowed patterns.
        """
        if self.is_key_allowed(key):
            super().__setitem__(key, value)

    def setdefault(self, key, default=None):
        """
        Set default value for a key if it matches allowed patterns.
        """
        if self.is_key_allowed(key):
            return super().setdefault(key, default)
        return None

    def update(self, mapping, **kwargs):
        """
        Update the dict with another mapping, only adding allowed keys.
        """
        assert not kwargs
        for key, value in mapping.items():
            if self.is_key_allowed(key):
                super().__setitem__(key, value)

    def is_key_allowed(self, key: str) -> bool:
        """
        Check if the key matches any of the allowed patterns.

        :param key:
        :return: True if the key is allowed, False otherwise.
        """
        if self.allowed_key_patterns is None:
            return True  # If no patterns defined, allow all keys
        for pattern in self.allowed_key_patterns:
            if fnmatch.fnmatch(key, pattern):
                return True
        return False


def is_key_allowed_in_collect_outputs_dict(collect_outputs: Union[CollectOutputsDict, dict], key: str) -> bool:
    """
    Check if a key is allowed in the given CollectOutputsDict.

    :param collect_outputs:
    :param key:
    :return: True if the key is allowed, False otherwise.
    """
    if isinstance(collect_outputs, CollectOutputsDict):
        return collect_outputs.is_key_allowed(key)
    return True  # If it's a regular dict, all keys are allowed



"""
Setups the environment for tests.
In the test code, you would have this::

    import _setup_test_env  # noqa

Also see :mod:`_setup_returnn_env`.
See :func:`setup` below for details.
"""

from __future__ import annotations


def setup():
    """
    Calls necessary setups.
    """
    import logging
    import os
    import sys

    os.environ["RETURNN_TEST"] = "1"

    # Enable all logging, up to debug level.
    logging.basicConfig(level=logging.DEBUG, format="%(message)s")

    # Disable extensive TF debug verbosity. Must come before the first TF import.
    # logging.getLogger("tensorflow").disabled = True
    # os.environ['TF_CPP_MIN_LOG_LEVEL'] = '0'
    # logging.getLogger("tensorflow").setLevel(logging.INFO)

    # Get us some further useful debug messages (in some cases, e.g. CUDA).
    # For example: https://github.com/tensorflow/tensorflow/issues/24496
    # os.environ["CUDNN_LOGINFO_DBG"] = "1"
    # os.environ["CUDNN_LOGDEST_DBG"] = "stdout"
    # The following might fix (workaround): Could not create cudnn handle: CUDNN_STATUS_INTERNAL_ERROR
    # (https://github.com/tensorflow/tensorflow/issues/24496).
    # os.environ["TF_FORCE_GPU_ALLOW_GROWTH"] = "true"

    import _setup_returnn_env  # noqa

    import returnn.util.basic as util

    # noinspection PyProtectedMember
    util.BehaviorVersion.set_min_behavior_version(util.BehaviorVersion._latest_behavior_version)

    from returnn.util import better_exchook

    if sys.excepthook != sys.__excepthook__:
        prev_sys_excepthook = sys.excepthook

        def _chained_excepthook(exctype, value, traceback):
            better_exchook.better_exchook(exctype, value, traceback)
            prev_sys_excepthook(exctype, value, traceback)

        sys.excepthook = _chained_excepthook
    else:
        sys.excepthook = better_exchook.better_exchook
    better_exchook.replace_traceback_format_tb()

    from returnn.log import log

    # No propagate, use stdout directly.
    log.initialize(verbosity=[5], propagate=False)

    import returnn.util.debug as debug

    debug.init_faulthandler()

    _try_hook_into_tests()


def _try_hook_into_tests():
    """
    Hook into nosetests or other unittest based frameworks.

    The hook will throw exceptions such that a debugger like PyCharm can inspect them easily.
    This will only be done if there is just a single test case.

    This code might be a bit experimental.
    It should work though. But if it does not, we can also skip this.
    Currently any exception here would be fatal though, as we expect this to work.

    Also see: https://youtrack.jetbrains.com/issue/PY-9848
    """
    import sys
    import types

    get_trace = getattr(sys, "gettrace", None)
    in_debugger = False
    if get_trace and get_trace() is not None:
        in_debugger = True

    # get TestProgram instance from stack...
    from unittest import TestProgram
    from returnn.util.better_exchook import get_current_frame
    from returnn.util.better_exchook import get_func_str_from_code_object

    top_frame = get_current_frame()
    if not top_frame:
        # This will not always work. Just silently accept this. This should be rare.
        return

    test_program = None
    frame = top_frame
    while frame:
        local_self = frame.f_locals.get("self")
        if isinstance(local_self, TestProgram):
            test_program = local_self
            break
        frame = frame.f_back

    test_names = None
    if test_program:  # nosetest, unittest
        test_names = getattr(test_program, "testNames")

    test_session = None
    try:
        # noinspection PyPackageRequirements,PyUnresolvedReferences
        import pytest
    except ImportError:
        pass
    else:
        frame = top_frame
        while frame:
            local_self = frame.f_locals.get("self")
            if isinstance(local_self, pytest.Session):
                test_session = local_self
                break
            frame = frame.f_back
        if test_session and not test_names:
            test_names = test_session.config.args

    if not test_names:
        # Unexpected, but just silently ignore.
        return
    if len(test_names) >= 2 or ":" not in test_names[0]:
        # Multiple tests are being run. Do not hook into this.
        # We only want to install the hook if there is only a single test case.
        return

    # Skip this if we are not in a debugger.
    if test_program and in_debugger:  # nosetest, unittest
        # Ok, try to install our plugin.
        class _ReraiseExceptionTestHookPlugin:
            @staticmethod
            def _reraise_exception(test, err):
                exc_class, exc, tb = err
                print("Test %s, exception %s %s, reraise now." % (test, exc_class.__name__, exc))
                raise exc

            handleFailure = _reraise_exception
            handleError = _reraise_exception

        config = getattr(test_program, "config")
        config.plugins.addPlugin(_ReraiseExceptionTestHookPlugin())

    if test_session and in_debugger:
        items = []

        def _custom_pytest_runtestloop(session):
            print("test env hook pytest_runtestloop")
            assert len(session.items) == len(test_names) == 1
            items.extend(session.items)

        def _custom_pytest_sessionfinish(session, exitstatus):
            session, exitstatus  # noqa  # not used
            print("test env hook pytest_sessionfinish")

        class _CustomPlugin:
            # noinspection PyShadowingNames
            def pytest_unconfigure(self, config):
                """hook for pytest_unconfigure."""
                print("test env hook pytest_unconfigure")
                # This will get called (potentially) multiple times
                # via config._ensure_unconfigure in the `finally` block
                # in certain stages of the pytest call stack.
                frame_ = get_current_frame()
                while frame_:
                    assert isinstance(frame_, types.FrameType)
                    # If pytest_cmdline_main is in the call stack trace, we are not yet in the lowest stack.
                    if get_func_str_from_code_object(frame_.f_code, frame=frame_) == "pytest_cmdline_main":
                        # We assume there is yet another lower `finally` block in the call stack
                        # which calls to config._ensure_unconfigure.
                        # However, it would not call pytest_unconfigure again
                        # when the configured flag is already set to False.
                        # So we again call config._do_configure to set the flag to True again.
                        # We don't want to run any of the other plugin hooks anymore,
                        # so unregister them all (except us).
                        for plugin in test_session.config.pluginmanager.get_plugins():
                            if plugin != self:
                                test_session.config.pluginmanager.unregister(plugin)
                        # noinspection PyProtectedMember
                        config._do_configure()  # causes this to run again
                        return
                    frame_ = frame_.f_back
                print("test env hook pytest_unconfigure, final call")
                config.add_cleanup(self._custom_final_cleanup)

            @staticmethod
            def _custom_final_cleanup():
                print("test env hook pytest config cleanup")
                print("Now calling the test:", items[0])
                items[0].obj()

        # Overwrite these hooks completely - we don't want to run the tests at this point,
        # as all exceptions would be caught here.
        test_session.config.hook.pytest_runtestloop = _custom_pytest_runtestloop
        test_session.config.hook.pytest_sessionfinish = _custom_pytest_sessionfinish
        # For other things, using a plugin to register the hooks is probably the more clean way.
        test_session.config.pluginmanager.register(_CustomPlugin())


setup()



"""
Helper to compile Torch ops on-the-fly, similar to Theano / :class:`returnn.tf.util.basic.OpCodeCompiler`,
similar to :mod:`torch.utils.cpp_extension`.

See :class:`OpCodeCompiler`.
"""

from __future__ import annotations
from typing import Union, Optional, Sequence, Dict, List
import os
import sysconfig

import torch
from torch.utils import cpp_extension

from returnn.util.basic import NativeCodeCompiler
from returnn.util.cuda_env import CudaEnv as _CudaEnvBase, get_best_nvcc_path_for_cuda_version


class OpCodeCompiler(NativeCodeCompiler):
    """
    Helper class to compile Torch ops on-the-fly, similar to Theano,
    and similar to :class:`returnn.tf.util.basic.OpCodeCompiler`.

    Note that PyTorch already has its own code for this,
    see :mod:`torch.utils.cpp_extension`, :func:`torch.utils.cpp_extension.load_inline`, etc.
    However, there are some shortcomings there that we try to do better:

    * The way we find CUDA/nvcc is more robust.
    * The way we find the C/C++ compiler is more robust.
    * The automatic selection of options for nvcc is more robust.
      E.g. the compute version is not higher than what the selected CUDA supports.

    https://docs.pytorch.org/tutorials/advanced/cpp_custom_ops.html
    """

    CacheDirName = "returnn_torch_cache/ops"

    def __init__(
        self,
        base_name: str,
        *,
        code: str,
        use_cuda_if_available: bool = True,
        cuda_auto_min_compute_capability: bool = True,
        include_paths: Sequence[str] = (),
        ld_flags: Sequence[str] = (),
        c_macro_defines: Optional[Dict[str, Union[str, int, None]]] = None,
        is_python_module: bool = False,
        **kwargs,
    ):
        self._cuda_env = None
        if use_cuda_if_available and torch.cuda.is_available():
            self._cuda_env = CudaEnv.get_instance()
            # Currently we assume that if we provide CUDA code (thus set use_cuda_if_available=True),
            # that if there is a GPU available (as TF reports it),
            # we also expect that we find CUDA.
            # Otherwise you would end up with ops compiled for CPU only although they support CUDA
            # and the user expects them to run on GPU.
            assert self._with_cuda(), "OpCodeCompiler: use_cuda_if_available=True but no CUDA found"

        self._nvcc_opts = []
        if self._with_cuda() and cuda_auto_min_compute_capability:
            # Get CUDA compute capability of the current GPU device.
            min_compute_capability = _get_available_gpu_cuda_min_compute_capability()
            if min_compute_capability:
                min_compute_capability = min(min_compute_capability, self._cuda_env.get_max_compute_capability())
                self._nvcc_opts += ["-arch", "compute_%i" % int(min_compute_capability * 10)]

        if self._with_cuda():
            self._nvcc_opts += cpp_extension.COMMON_NVCC_FLAGS

        # Example call from torch.utils.cpp_extension:
        # /usr/local/cuda-11.0/bin/nvcc
        # --generate-dependencies-with-compile --dependency-output cuda.cuda.o.d
        # -DTORCH_EXTENSION_NAME=async_assert_ext
        # -DTORCH_API_INCLUDE_EXTENSION_H
        # -isystem /home/az/py-venv/py3.12-torch2.9/lib/python3.12/site-packages/torch/include
        # -isystem /home/az/py-venv/py3.12-torch2.9/lib/python3.12/site-packages/torch/include/torch/csrc/api/include
        # -isystem /usr/local/cuda-11.0/include -isystem /usr/include/python3.12
        # -D__CUDA_NO_HALF_OPERATORS__ -D__CUDA_NO_HALF_CONVERSIONS__
        # -D__CUDA_NO_BFLOAT16_CONVERSIONS__ -D__CUDA_NO_HALF2_OPERATORS__
        # --expt-relaxed-constexpr
        # -gencode=arch=compute_86,code=compute_86 -gencode=arch=compute_86,code=sm_86
        # --compiler-options '-fPIC' -std=c++17
        # -c /home/az/.cache/torch_extensions/py312_cu128/async_assert_ext/cuda.cu -o cuda.cuda.o

        torch_path = os.path.dirname(torch.__file__)
        torch_include = torch_path + "/include"
        assert os.path.isdir(torch_include)

        python_include = sysconfig.get_path("include", scheme="posix_prefix")

        include_paths = list(include_paths) + [torch_include, torch_include + "/torch/csrc/api/include", python_include]

        c_macro_defines = {} if c_macro_defines is None else c_macro_defines.copy()
        c_macro_defines.setdefault("TORCH_EXTENSION_NAME", base_name)
        c_macro_defines.setdefault("TORCH_API_INCLUDE_EXTENSION_H", "")
        # We have some assert in our kernels that we want to disable.
        c_macro_defines.setdefault("NDEBUG", 1)

        ld_flags = list(ld_flags)
        ld_flags.append("--no-as-needed")
        ld_flags.append(f"-L{cpp_extension.TORCH_LIB_PATH}")
        ld_flags.append("-lc10")
        if self._with_cuda():
            ld_flags.append("-lc10_cuda")
        ld_flags.append("-ltorch_cpu")
        if self._with_cuda():
            ld_flags.append("-ltorch_cuda")
        ld_flags.append("-ltorch")
        ld_flags.append("-ltorch_python")

        if self._with_cuda():
            ld_flags.append(self._cuda_env.get_ld_flag_for_linking_cudart())
            # maybe add CUDNN?

        # noinspection PyUnresolvedReferences,PyProtectedMember
        use_cxx11_abi = torch._C._GLIBCXX_USE_CXX11_ABI

        super().__init__(
            base_name=base_name,
            code=code,
            include_paths=include_paths,
            c_macro_defines=c_macro_defines,
            ld_flags=ld_flags,
            use_cxx11_abi=use_cxx11_abi,
            **kwargs,
        )
        self.is_python_module = is_python_module
        self._mod = None

    def __repr__(self):
        return "<%s %r CUDA %s in %r>" % (self.__class__.__name__, self.base_name, self._with_cuda(), self._mod_path)

    _relevant_info_keys = NativeCodeCompiler._relevant_info_keys + (
        "torch_version",
        "with_cuda",
        "cuda_path",
        "nvcc_opts",
    )

    def _make_info_dict(self):
        from returnn.util.basic import describe_torch_version

        d = super()._make_info_dict()
        d.update(
            {
                "torch_version": describe_torch_version(),
                "with_cuda": self._with_cuda(),
                "cuda_path": self._cuda_env.cuda_path if self._with_cuda() else None,
                "nvcc_opts": (
                    (tuple(self._cuda_env.get_compiler_opts()) + tuple(self._nvcc_opts)) if self._with_cuda() else None
                ),
            }
        )
        return d

    @classmethod
    def cuda_available(cls):
        """
        :return: whether CUDA is available. if True, and you initiate with use_cuda_if_available=True,
          then _with_cuda() should also be True.
        :rtype: bool
        """
        if not torch.cuda.is_available():
            return False
        cuda_env = CudaEnv.get_instance()
        return cuda_env.is_available()

    def _with_cuda(self):
        return bool(self._cuda_env and self._cuda_env.is_available())

    cpp_version = 17

    def _get_compiler_bin(self):
        if self._with_cuda():
            return self._cuda_env.get_compiler_bin()
        return super()._get_compiler_bin()

    def _transform_compiler_opts(self, opts: List[str]) -> List[str]:
        if self._with_cuda():
            nvcc_opts = self._cuda_env.get_compiler_opts()
            for opt in opts:
                nvcc_opts += ["-Xcompiler", opt]
            nvcc_opts += self._nvcc_opts
            return nvcc_opts
        return super()._transform_compiler_opts(opts)

    def _transform_ld_flags(self, opts: Sequence[str]) -> Sequence[str]:
        if self._with_cuda():
            res = []
            for opt in opts:
                if opt.startswith("-L") or opt.startswith("-l"):
                    res.append(opt)
                else:
                    res += ["-Xlinker", opt]
            return res
        return super()._transform_ld_flags(opts)

    def load_module(self):
        """
        :return: module
        """
        if self._mod:
            return self._mod
        self._maybe_compile()

        if self.is_python_module:
            # Load as a Python module.
            # E.g. PYBIND11_MODULE or so was used in the code.
            import importlib.util

            # https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
            spec = importlib.util.spec_from_file_location(self.base_name, self._so_filename)
            assert spec is not None
            module = importlib.util.module_from_spec(spec)
            assert isinstance(spec.loader, importlib.abc.Loader)
            spec.loader.exec_module(module)

        else:
            # Load as a Torch extension.
            # TORCH_LIBRARY / TORCH_LIBRARY_IMPL was used in the code.
            torch.ops.load_library(self._so_filename)
            module = getattr(torch.ops, self.base_name)

        self._mod = module
        return module


class CudaEnv(_CudaEnvBase):
    """specialized CudaEnv for PyTorch"""

    # If cudart is loaded (e.g. via Torch), we really want to use that one.
    _runtime_libcudart_path_must_be_valid = True

    def __init__(self):
        super().__init__()

        from returnn.util.basic import find_libcudart_from_runtime

        self._runtime_libcudart = find_libcudart_from_runtime()
        self._compiler_bin = None
        if self.cuda_path:
            if os.path.exists(f"{self.cuda_path}/bin/nvcc"):
                self._compiler_bin = f"{self.cuda_path}/bin/nvcc"
            else:
                self._compiler_bin = get_best_nvcc_path_for_cuda_version(self.get_cuda_version())

    @classmethod
    def _check_valid_cuda_path(cls, p: str) -> bool:
        """
        :param p: path to CUDA, e.g. "/usr/local/cuda-8.0"
        :return: whether this is a valid CUDA path, i.e. we find all what we need
        """
        if cls.verbose_find_cuda:
            print("check valid CUDA path: %s" % p)
        # Don't check nvcc here yet.
        # The pip package might not have it, but otherwise provides lib + headers
        # that we want to use, as this is likely the same that PyTorch uses.
        if not os.path.exists("%s/include/cuda.h" % p):
            return False
        if p.endswith("/site-packages/nvidia/cuda_runtime"):
            # special case for the nvidia CUDA pip package
            if not any(name.startswith("libcudart.") for name in os.listdir(p + "/lib")):
                return False
        else:
            if not os.path.exists("%s/%s/libcudart.so" % (p, cls._get_lib_dir_name(p))):
                return False
        return True

    def get_lib_dir_path(self) -> str:
        """
        :return: path
        """
        if self._runtime_libcudart:
            return os.path.dirname(self._runtime_libcudart)
        return super().get_lib_dir_path()

    def get_ld_flag_for_linking_cudart(self) -> str:
        """ld flag"""
        if self._runtime_libcudart:
            return f"-l:{os.path.basename(self._runtime_libcudart)}"
        return "-lcudart"

    def get_compiler_bin(self) -> str:
        """
        :return: path
        """
        return self._compiler_bin


def _get_available_gpu_cuda_min_compute_capability() -> Optional[float]:
    """
    Uses :func:`get_available_gpu_devices`.

    :return: e.g. 3.0, or 5.0, etc, or None
    """
    count = torch.cuda.device_count()
    cap = None
    for i in range(count):
        props = torch.cuda.get_device_properties(i)
        dev_cap = float(f"{props.major}.{props.minor}")
        if cap is None:
            cap = dev_cap
        else:
            cap = min(cap, dev_cap)
    return cap



"""
CUDA environment detection and information.
"""

from __future__ import annotations
from typing import Dict, Tuple, List
import os
import re


class CudaEnv:
    """
    Information about the Nvidia CUDA environment, and library.
    Also path to ``nvcc``, the CUDA compiler.
    """

    _instance_per_cls: Dict[type, CudaEnv] = {}
    verbose_find_cuda = False

    def __init__(self):
        from returnn.util.basic import to_bool

        if to_bool(os.environ.get("DISABLE_CUDA", "0")):
            self.cuda_path = None
            if self.verbose_find_cuda:
                print("CUDA disabled via env DISABLE_CUDA.")
        elif os.environ.get("CUDA_VISIBLE_DEVICES", None) in ["", "-1"]:
            self.cuda_path = None
            if self.verbose_find_cuda:
                print(f"CUDA disabled via env CUDA_VISIBLE_DEVICES={os.environ['CUDA_VISIBLE_DEVICES']!r}.")
        else:
            self.cuda_path = self._find_cuda_path()
            if self.verbose_find_cuda:
                print("CUDA path:", self.cuda_path)

        self._max_compute_capability = None
        self._cuda_version = None

    @classmethod
    def _find_nvcc_in_path(cls):
        """
        :return: yields full path to nvcc
        :rtype: list[str]
        """
        for p in os.environ["PATH"].split(":"):
            pp = "%s/nvcc" % p
            if os.path.exists(pp):
                yield pp

    @classmethod
    def _find_lib_in_ld_path(cls):
        """
        :return: yields full path to libcudart.so
        :rtype: list[str]
        """
        from returnn.util.basic import get_ld_paths

        for p in get_ld_paths():
            pp = "%s/libcudart.so" % p
            if os.path.exists(pp):
                yield pp

    @classmethod
    def _get_lib_dir_name(cls, base_path):
        """
        :return: dir name in base path
        :rtype: str
        """
        from returnn.util.basic import is_64bit_platform, get_ld_paths

        for ld_path in get_ld_paths():
            # We also want to allow "lib/x86_64-linux-gnu" for "/usr".
            # However, this logic should not be triggered for incorrect cases.
            # E.g. base_path="/usr" would be the prefix for most LD paths.
            if ld_path.startswith(base_path + "/lib") and os.path.exists("%s/libcudart.so" % ld_path):
                return ld_path[len(base_path) + 1 :]
        if is_64bit_platform():
            return "lib64"
        return "lib"

    _runtime_libcudart_path_must_be_valid: bool = False

    @classmethod
    def _cuda_path_candidate_via_proc_map_libcudart(cls):
        from returnn.util.basic import find_libcudart_from_runtime

        fn = find_libcudart_from_runtime()
        if cls.verbose_find_cuda:
            print("libcudart.so found from /proc/maps:", fn)
        if not fn:
            return None
        # fn is e.g. '/usr/local/cuda-8.0/targets/x86_64-linux/lib/libcudart.so.8.0.61',
        # or maybe '/usr/local/cuda-8.0/lib64/libcudart.so'
        # or maybe ".../site-packages/nvidia/cuda_runtime/lib/libcudart.so.12"
        # or ".../site-packages/nvidia/cu13/lib/libcudart.so.13"
        p = os.path.dirname(os.path.dirname(fn))
        while not cls._check_valid_cuda_path(p):
            p = os.path.dirname(p)
            if p in ["", "/"]:
                if cls.verbose_find_cuda:
                    print(f"Loaded lib {fn} does not seem to be in valid CUDA path.")
                assert not cls._runtime_libcudart_path_must_be_valid
                return None
        assert cls._check_valid_cuda_path(p)
        return p

    @classmethod
    def _cuda_path_candidates(cls):
        p = cls._cuda_path_candidate_via_proc_map_libcudart()
        if p:
            yield p
        if os.environ.get("CUDA_HOME"):
            yield os.environ.get("CUDA_HOME")
        if os.environ.get("CUDA_PATH"):
            yield os.environ.get("CUDA_PATH")
        for p in cls._find_nvcc_in_path():
            # Expect p == "/usr/local/cuda-8.0/bin/nvcc" or so.
            postfix = "/bin/nvcc"
            if cls.verbose_find_cuda:
                print("found cuda nvcc (wanted postfix: %r): %s" % (postfix, p))
            if not p.endswith(postfix):
                continue
            yield p[: -len(postfix)] or "/"
        for p in cls._find_lib_in_ld_path():
            # Expect p == "/usr/local/cuda-8.0/lib64/libcudart.so" or so.
            d = "/".join(p.split("/")[:-2]) or "/"  # Get "/usr/local/cuda-8.0".
            if cls.verbose_find_cuda:
                print("found cuda lib: %s (path %s)" % (p, d))
            yield d
        # Check common installation location.
        for path in get_cuda_path_candidates_from_common_install_locations():
            yield path

    @classmethod
    def _check_valid_cuda_path(cls, p):
        """
        :param str p: path to CUDA, e.g. "/usr/local/cuda-8.0"
        :return: whether this is a valid CUDA path, i.e. we find all what we need
        :rtype: bool
        """
        if cls.verbose_find_cuda:
            print("check valid CUDA path: %s" % p)
        if not os.path.exists("%s/bin/nvcc" % p):
            return False
        if not os.path.exists("%s/include/cuda.h" % p):
            return False
        if not os.path.exists("%s/%s/libcudart.so" % (p, cls._get_lib_dir_name(p))):
            return False
        return True

    @classmethod
    def _find_cuda_path(cls):
        """
        :return: base CUDA path if we find one, otherwise None
        :rtype: str|None
        """
        for p in cls._cuda_path_candidates():
            if cls._check_valid_cuda_path(p):
                return p
        return None

    def is_available(self):
        """
        :rtype: bool
        """
        return bool(self.cuda_path)

    def get_cuda_version(self) -> Tuple[int, int]:
        """
        Get CUDA version as (major, minor).
        """
        if self._cuda_version:
            return self._cuda_version
        assert self.cuda_path
        # Parse CUDA_VERSION from cuda.h.
        cuda_h_path = f"{self.cuda_path}/include/cuda.h"
        self._cuda_version = _parse_cuda_version_from_cuda_h(cuda_h_path)
        return self._cuda_version

    def get_max_compute_capability(self):
        """
        :return: the highest compute capability supported by nvcc, or float("inf") if not known
        :rtype: float
        """
        if self._max_compute_capability is None:
            cuda_occupancy_path = "%s/include/cuda_occupancy.h" % self.cuda_path
            if os.path.exists(cuda_occupancy_path):
                major, minor = None, 0
                for line in open(cuda_occupancy_path).read().splitlines():
                    m = re.match("^#define\\s+__CUDA_OCC_(MAJOR|MINOR)__\\s+([0-9]+)$", line)
                    if m:
                        s, v = m.groups()
                        v = int(v)
                        if s == "MAJOR":
                            major = v
                        else:
                            minor = v
                if major:
                    self._max_compute_capability = float(major) + float(minor) * 0.1
        if self._max_compute_capability is None:
            self._max_compute_capability = float("inf")
        return self._max_compute_capability

    @staticmethod
    def get_cc_bin() -> str:
        """
        :return: path
        """
        from .native_code_compiler import get_cc_bin

        return get_cc_bin()

    def get_compiler_opts(self):
        """
        :rtype: list[str]
        """
        return [
            "-ccbin",
            self.get_cc_bin(),
            "-I",
            "%s/targets/x86_64-linux/include" % self.cuda_path,
            "-I",
            "%s/include" % self.cuda_path,
            "-L",
            self.get_lib_dir_path(),
            "-x",
            "cu",
            "-v",
        ]

    def get_lib_dir_path(self) -> str:
        """library path"""
        return "%s/%s" % (self.cuda_path, self._get_lib_dir_name(self.cuda_path))

    def get_compiler_bin(self):
        """
        :return: path
        :rtype: str
        """
        assert self.cuda_path
        return "%s/bin/nvcc" % self.cuda_path

    @classmethod
    def get_instance(cls) -> CudaEnv:
        """
        :return: instance for this class
        """
        if cls._instance_per_cls.get(cls) is not None:
            return cls._instance_per_cls[cls]
        cls._instance_per_cls[cls] = cls()
        return cls._instance_per_cls[cls]


def get_cuda_path_candidates_from_common_install_locations() -> List[str]:
    """
    :return: list of possible CUDA installation paths from common locations
    """
    cuda_paths = []

    if os.path.exists("/usr/local"):
        for name in sorted(os.listdir("/usr/local")):
            if name.startswith("cuda-") or name == "cuda":
                p = f"/usr/local/{name}"
                if _check_valid_cuda_path_with_nvcc(p):
                    version = _parse_cuda_version_from_cuda_h(f"{p}/include/cuda.h")
                    cuda_paths.append((version, p))

    # (stable) sort by version, highest version first
    cuda_paths.sort(key=lambda x: x[0], reverse=True)
    return [p for (_, p) in cuda_paths]


def get_best_nvcc_path_for_cuda_version(cuda_version: Tuple[int, int]) -> str:
    """
    :return: path to nvcc
    :rtype: str
    """
    cuda_paths = []

    # noinspection PyProtectedMember
    for p in CudaEnv._cuda_path_candidates():
        if _check_valid_cuda_path_with_nvcc(p):
            version = _parse_cuda_version_from_cuda_h(f"{p}/include/cuda.h")
            if version == cuda_version:
                # if we found a matching one, directly return it
                return f"{p}/bin/nvcc"
            cuda_paths.append((version, p))

    if not cuda_paths:
        raise RuntimeError(f"No valid CUDA installation found for version {cuda_version}.")

    only_higher_versions = [(version, p) for (version, p) in cuda_paths if version >= cuda_version]
    if only_higher_versions:
        only_higher_versions.sort(key=lambda x: x[0])
        # return the lowest higher version
        if only_higher_versions[0][0] != cuda_version[0]:  # major version differs
            print(
                f"Warning: No exact match for CUDA version {cuda_version}, "
                f"using version {only_higher_versions[0]} instead."
            )
        return f"{only_higher_versions[0][1]}/bin/nvcc"

    cuda_paths.sort(key=lambda x: x[0])
    # return the highest lower version
    print(f"Warning: No exact match for CUDA version {cuda_version}, using lower version {cuda_paths[-1][0]} instead.")
    return f"{cuda_paths[-1][1]}/bin/nvcc"


def _check_valid_cuda_path_with_nvcc(p: str) -> bool:
    """
    :param str p: path to CUDA, e.g. "/usr/local/cuda-8.0"
    :return: whether this is a valid CUDA path, i.e. we find all what we need
    :rtype: bool
    """
    if not os.path.exists("%s/bin/nvcc" % p):
        return False
    if not os.path.exists("%s/include/cuda.h" % p):
        return False
    return True


def _parse_cuda_version_from_cuda_h(cuda_h_path: str) -> Tuple[int, int]:
    assert os.path.exists(cuda_h_path)
    for line in open(cuda_h_path).read().splitlines():
        # Like: #define CUDA_VERSION 12080
        m = re.match(r"^#define\s+CUDA_VERSION\s+([0-9]+)$", line)
        if m:
            version_num = int(m.group(1))
            major = version_num // 1000
            minor = (version_num % 1000) // 10
            return major, minor
    raise RuntimeError(f"Could not determine CUDA version from {cuda_h_path}.")



"""
Generic framework independent reference implementations
"""

from typing import Any, Optional, Union, Dict, List, Tuple
from collections import defaultdict
import numpy


NEG_INF = -float("inf")


def logsumexp(x: Union[numpy.ndarray, Any]) -> float:
    """
    Stable log sum exp.
    """
    x = numpy.array(x)
    if (x == NEG_INF).all():
        return NEG_INF
    a_max = x.max()
    lsp = numpy.log(numpy.sum(numpy.exp(x - a_max)))
    return a_max + lsp


def py_baum_welch(am_scores, float_idx, edges, weights, start_end_states):
    """
    Pure Python Forward-backward (Baum Welch) algorithm.
    The parameters are in the same format as our native fast_baum_welch op.

    :param numpy.ndarray am_scores: (time, batch, dim), in -log space
    :param numpy.ndarray float_idx: (time, batch) -> 0 or 1 (index mask, via seq lens)
    :param numpy.ndarray edges: (4,num_edges), edges of the graph (from,to,emission_idx,sequence_idx)
    :param numpy.ndarray weights: (num_edges,), weights of the edges
    :param numpy.ndarray start_end_states: (2, batch), (start,end) state idx in FSA. there is only one single FSA.
    :return: (fwdbwd, obs_scores), fwdbwd is (time, batch, dim), obs_scores is (time, batch), in -log space
    :rtype: (numpy.ndarray, numpy.ndarray)
    """
    # We get it in -log space, but we calculate in +log space.
    am_scores = -am_scores
    weights = -weights
    n_time, n_batch, dim = am_scores.shape
    assert float_idx.shape == (n_time, n_batch)
    assert edges.ndim == 2 and weights.ndim == 1
    (n_edges,) = weights.shape
    assert edges.shape == (4, n_edges)
    assert start_end_states.shape == (2, n_batch)

    zero_score = float("-inf")
    fwdbwd = numpy.zeros((n_time, n_batch, dim), dtype=am_scores.dtype) + zero_score
    obs_scores = numpy.zeros((n_time, n_batch), dtype=am_scores.dtype) + zero_score

    def collect_scores(forward: bool) -> List[Dict[int, float]]:
        """
        :param forward:
        """
        start_idx, end_idx = start_end_states[:, sequence_idx]
        states: Dict[int, float] = defaultdict(lambda: zero_score)  # state-idx -> score.
        states[start_idx if forward else end_idx] = 0.0
        scores_over_t: List[Optional[Dict[int, float]]] = [None] * (n_time + 1)
        scores_over_t[0 if forward else -1] = dict(states)
        for t in range(n_time) if forward else reversed(range(n_time)):
            if float_idx[t, sequence_idx] == 1:
                scores: Dict[int, List[float]] = defaultdict(list)  # state-idx -> list[score]
                for edge_idx in range(n_edges):
                    from_idx, to_idx, emission_idx, sequence_idx_ = edges[:, edge_idx]
                    if not forward:
                        from_idx, to_idx = to_idx, from_idx
                    if sequence_idx_ != sequence_idx:
                        continue
                    if from_idx not in states or states[from_idx] == zero_score:
                        continue
                    assert 0 <= emission_idx < dim
                    score = states[from_idx] + weights[edge_idx] + am_scores[t, sequence_idx, emission_idx]
                    scores[to_idx].append(score)
                states.clear()
                for state_idx in scores.keys():
                    states[state_idx] = float(logsumexp(scores[state_idx]))
            scores_over_t[(t + 1) if forward else t] = dict(states)
        return scores_over_t

    def gamma():
        """
        :return: nothing, fill fwdbwd and obs_scores
        """
        for t in range(n_time):
            if float_idx[t, sequence_idx] == 1:
                scores: Dict[int, List[float]] = defaultdict(list)  # emission-idx -> list[score]
                all_scores: List[float] = []
                for edge_idx in range(n_edges):
                    from_idx, to_idx, emission_idx, sequence_idx_ = edges[:, edge_idx]
                    if sequence_idx_ != sequence_idx:
                        continue
                    assert 0 <= emission_idx < dim
                    if from_idx not in fwd_scores[t]:
                        continue
                    if to_idx not in bwd_scores[t + 1]:
                        continue
                    score = (
                        fwd_scores[t][from_idx]
                        + weights[edge_idx]
                        + am_scores[t, sequence_idx, emission_idx]
                        + bwd_scores[t + 1][to_idx]
                    )
                    scores[emission_idx].append(score)
                    all_scores.append(score)
                obs_scores[t, sequence_idx] = logsumexp(all_scores) if all_scores else zero_score
                for emission_idx, values in scores.items():
                    if not values:
                        fwdbwd[t, sequence_idx, emission_idx] = zero_score
                    else:
                        fwdbwd[t, sequence_idx, emission_idx] = float(logsumexp(values)) - obs_scores[t, sequence_idx]

    for sequence_idx in range(n_batch):
        fwd_scores = collect_scores(forward=True)
        bwd_scores = collect_scores(forward=False)
        gamma()

    # -log space
    return -fwdbwd, -obs_scores


def py_viterbi(am_scores, am_seq_len, edges, weights, start_end_states):
    """
    Pure Python Viterbi algorithm, to find the best path/alignment.
    The parameters are in the same format as our native fast_baum_welch op.

    :param numpy.ndarray am_scores: (time, batch, dim), in +log space
    :param numpy.ndarray am_seq_len: (batch,) -> int32
    :param numpy.ndarray edges: (4,num_edges), edges of the graph (from,to,emission_idx,sequence_idx)
    :param numpy.ndarray weights: (num_edges,), weights of the edges
    :param numpy.ndarray start_end_states: (2, batch), (start,end) state idx in FSA. there is only one single FSA.
    :return: (alignment, obs_scores), alignment is (time, batch), obs_scores is (batch,), in +log space
    :rtype: (numpy.ndarray, numpy.ndarray)
    """
    n_time, n_batch, dim = am_scores.shape
    assert am_seq_len.shape == (n_batch,)
    assert edges.ndim == 2 and weights.ndim == 1
    (n_edges,) = weights.shape
    assert edges.shape == (4, n_edges)
    assert start_end_states.shape == (2, n_batch)

    zero_score = float("-inf")
    alignment = numpy.zeros((n_time, n_batch), dtype="int32")
    obs_scores = numpy.zeros((n_batch,), dtype=am_scores.dtype) + zero_score

    def search():
        """
        :rtype: list[dict[int,(float,int)]]
        """
        start_idx, _ = start_end_states[:, sequence_idx]
        states: Dict[int, Tuple[float, int]] = defaultdict(lambda: (zero_score, -1))  # state-idx -> score/edge
        states[start_idx] = (0.0, -1)
        res: List[Dict[int, Tuple[float, int]]] = []
        for t in range(n_time):
            if t >= am_seq_len[sequence_idx]:
                break
            scores: Dict[int, List[Tuple[float, int]]] = defaultdict(list)  # state-idx -> list[score/edge]
            for edge_idx in range(n_edges):
                from_idx, to_idx, emission_idx, sequence_idx_ = edges[:, edge_idx]
                if sequence_idx_ != sequence_idx:
                    continue
                if from_idx not in states or states[from_idx] == zero_score:
                    continue
                assert 0 <= emission_idx < dim
                score = states[from_idx][0] + weights[edge_idx] + am_scores[t, sequence_idx, emission_idx]
                scores[to_idx].append((score, edge_idx))
            states.clear()
            for state_idx in scores.keys():
                states[state_idx] = max(scores[state_idx], key=lambda _item: (_item[0], -_item[1]))
            res.append(dict(states))
        assert len(res) == am_seq_len[sequence_idx]
        return res

    def select_best():
        """
        :return: nothing, fill alignment and obs_scores
        """
        _, end_idx = start_end_states[:, sequence_idx]
        state_idx = end_idx
        for t in reversed(range(am_seq_len[sequence_idx])):
            if state_idx not in fwd_search_res[t]:  # no path?
                alignment[t, sequence_idx] = 0
                continue
            score, edge_idx = fwd_search_res[t][state_idx]
            if t == am_seq_len[sequence_idx] - 1:
                obs_scores[sequence_idx] = score
            from_idx, to_idx, emission_idx, sequence_idx_ = edges[:, edge_idx]
            assert sequence_idx_ == sequence_idx
            alignment[t, sequence_idx] = emission_idx
            state_idx = from_idx

    for sequence_idx in range(n_batch):
        fwd_search_res = search()
        select_best()

    return alignment, obs_scores



"""
Code to create PyTorch datasets that can be used with the PyTorch DataLoader.

We make use of torch.utils.data.IterDataPipe data pipelines.
(We used TorchData before but migrated back to pure PyTorch. https://github.com/rwth-i6/returnn/issues/1382)

Most functionality is implemented as a dataset/datapipe, as this seems to be the common way in PyTorch,
as it is also commonly done in Fairseq:
    https://github.com/facebookresearch/fairseq/tree/main/fairseq/data
    https://github.com/facebookresearch/fairseq/blob/main/fairseq/data/subsample_dataset.py

This is also the intended way for TorchData.

We potentially could also implement some functionality as part of the data loader (v1),
but DataLoader2 suggests to decouple this, as we do here.

We also have :class:`ChunkShuffleDataset` on RETURNN dataset level.
However, having this separate pure PyTorch implementation is useful to allow to use
other PyTorch datasets more directly, including also HuggingFace datasets.
"""

from __future__ import annotations
import bisect
import itertools
from typing import Optional, Any, Sequence, Tuple, Union, List, Dict, Callable
import sys
from copy import deepcopy

import numpy
import torch
import torch.utils.data

from returnn.config import Config
from returnn.log import log
from returnn.util.basic import NumbersDict, get_fwd_compat_kwargs


def create_tensor(array: numpy.ndarray) -> Union[torch.Tensor, numpy.ndarray]:
    """
    Adjust non-supported dtypes

    :param array: numpy array to be converted
    """
    # The only supported PyTorch dtypes are:
    # float64, float32, float16, complex64, complex128, int64, int32, int16, int8, uint8, and bool.
    if array.dtype.kind in "UO":  # string (unicode) or object
        return array  # keep as-is. e.g. seq_tag
    if array.dtype == numpy.uint32:
        array = numpy.asarray(array, dtype=numpy.int64)
    elif array.dtype == numpy.uint16:
        array = numpy.asarray(array, dtype=numpy.int32)
    return torch.tensor(array)


def collate_batch(batch: List[Dict[str, numpy.ndarray]]) -> Dict[str, Union[torch.Tensor, numpy.ndarray]]:
    """
    :param batch:
    """
    assert isinstance(batch, list)
    assert batch, "batch is empty?"
    assert isinstance(batch[0], dict)
    data_keys = list(batch[0].keys())

    res = {}
    for key in data_keys:
        if key in ("num_seqs", "epoch"):
            res[key] = batch[0][key]  # it should always be the same
            continue
        elif key == "complete_frac":
            res[key] = max(sample[key] for sample in batch)
            continue
        ls = [create_tensor(sample[key]) for sample in batch]
        if not ls:
            raise ValueError("batch is empty?")
        if isinstance(ls[0], torch.Tensor):
            if ls[0].ndim > 0:
                padded = torch.nn.utils.rnn.pad_sequence(ls, batch_first=True, padding_value=0)
                res[key] = padded
                res["%s:seq_len" % key] = torch.tensor([v.shape[0] for v in ls], dtype=torch.int32)
            else:
                res[key] = torch.stack(ls, dim=0)
        elif isinstance(ls[0], numpy.ndarray):
            padded = numpy.stack(ls, axis=0)
            res[key] = padded

    return res


class ChunkingIterDataPipe(torch.utils.data.IterDataPipe):
    """
    Splits each sequence in the given dataset into chunks according to the 'chunking' config option.
    So it transforms one sequences into multiple sequences.
    """

    def __init__(self, dataset: torch.utils.data.IterableDataset, chunking, *, min_chunk_size=0):
        """
        :param dataset: dataset to apply chunking to
        :param None|int|(int,int)|dict|(dict,dict) chunking: tuple (chunk_size, chunk_step).
            If given as single value,
            value will be used for both.
            Both chunk_size and chunk_step can be given as a dict data_key -> size/step.
            This can be used to apply chunking to only a subset of all data keys,
            or to use different chunking for different
            data keys.
            (The number of resulting chunks has to be match though for all given data keys, i.e. sequence lengths
            have to be considered.)
        """
        super().__init__()
        from returnn.datasets.basic import Dataset as ReturnnDataset

        self._dataset = dataset
        # noinspection PyProtectedMember
        self._chunk_size, self._chunk_step, custom_chunk_func = ReturnnDataset._parse_chunking(chunking)
        self._min_chunk_size = NumbersDict(min_chunk_size)

        assert not custom_chunk_func, f"Custom chunking function not supported, {chunking!r}"

    def __iter__(self):
        """
        :return: generator providing chunks in the form of a dict data_key -> data chunk
        :rtype: Iterable[dict[str, numpy.ndarray]]
        """
        chunking_data_keys = list(self._chunk_size.keys())

        for data_dict in self._dataset:
            if not chunking_data_keys:
                chunking_data_keys = list(data_dict.keys())  # use all if not configured separately
                chunking_data_key_black_list = ["seq_tag", "seq_idx", "num_seqs", "epoch", "complete_frac"]
                for key in chunking_data_key_black_list:
                    if key in chunking_data_keys:
                        chunking_data_keys.remove(key)
                assert chunking_data_keys, "Dataset produced sequence without any data."

            data_chunks = {}
            num_chunks = None

            for data_key in chunking_data_keys:
                chunk_size = self._chunk_size[data_key]
                chunk_step = self._chunk_step[data_key]
                min_chunk_size = self._min_chunk_size[data_key]

                data = data_dict[data_key]
                chunks = [
                    data[start_index : start_index + chunk_size]
                    for start_index in range(0, len(data), chunk_step)
                    if len(data[start_index : start_index + chunk_size]) >= min_chunk_size
                ]

                if num_chunks is None:
                    num_chunks = len(chunks)
                else:
                    assert num_chunks == len(chunks), (
                        "Chunking resulted in different number of chunks for different data keys."
                    )

                data_chunks[data_key] = chunks

            if num_chunks == 0:
                continue
            assert num_chunks, "Bug: no chunk produced from current sequence."
            for chunk_index in range(num_chunks):
                chunk_data = {data_key: data_chunks[data_key][chunk_index] for data_key in data_chunks.keys()}

                # If chunking is configured using a dict,
                # i.e. with explicit data keys, there might be remaining data keys
                # for which we yield the full sequence in each chunk.
                non_chunked_data = {
                    data_key: data for data_key, data in data_dict.items() if data_key not in chunk_data
                }
                if non_chunked_data:
                    chunk_data.update(deepcopy(non_chunked_data))

                yield chunk_data

    def __getitem__(self, index):
        raise Exception(f"{self.__class__.__name__}.__getitem__ not supported")

    @staticmethod
    def _parse_chunking(chunking):
        """
        Similar to returnn.datasets.basic.Dataset._parse_chunking().

        :param None|int|(int,int)|dict|(dict,dict) chunking: see __init__()
        :return: chunk_size, chunk_step
        :rtype: (NumbersDict,NumbersDict)
        """
        if not isinstance(chunking, (tuple, list)):
            chunking = (chunking, None)
        chunk_size, chunk_step = chunking
        if chunk_size is None:
            chunk_size = 0
        assert isinstance(chunk_size, (int, dict))
        chunk_size = NumbersDict(chunk_size)
        assert chunk_size.min_value() > 0, "chunk size must not be negative"
        if chunk_step in (None, 0):
            chunk_step = chunk_size
        assert isinstance(chunk_step, (int, dict, NumbersDict))
        chunk_step = NumbersDict(chunk_step)
        assert sorted(chunk_step.keys()) == sorted(chunk_size.keys())
        assert chunk_step.min_value() > 0, "chunking step must be positive"
        return chunk_size, chunk_step


# noinspection PyAbstractClass
class BatchingIterDataPipe(torch.utils.data.IterDataPipe):
    """
    Converts a dataset yielding sequences (dict data_key -> array per sequence) into a dataset yielding lists of
    these sequences, i.e. batches.
    Sequences are grouped in-order according to the 'max_tokens' and 'max_seqs' batch size
    limits.
    Note, that batches are not yet merged into a single (padded) data array here, this happens in 'collate_batch()'.
    """

    def __init__(self, dataset: torch.utils.data.IterableDataset, batch_size=1, max_seqs=None):
        """
        :param dataset: dataset to apply batching to
        :param int|dict[str,int]|None|function batch_size: Maximum number of time steps (e.g. audio frames / words)
            in one batch (padding included).
            If given as a dict data_key -> value, sets different individual limits per data key.
            If None, no limit.
            Can also be a callable with kwargs epoch, seq_idx, epoch_continuous, **_other_kwargs,
            returning the batch size.
        :param int|None|function max_seqs: maximum number of sequences in a batch,
            None means unlimited (also -1 to match TF backend).
            Can also be a callable with kwargs epoch, seq_idx, epoch_continuous, **_other_kwargs,
            returning the max seqs.
        """
        super().__init__()
        self._dataset = dataset
        self._max_batch_size = self._parse_batch_size(batch_size)
        self._max_seqs = self._parse_max_seqs(max_seqs)

        if not callable(self._max_batch_size):
            assert isinstance(self._max_batch_size, NumbersDict) and self._max_batch_size.min_value() > 0
        if not callable(self._max_seqs):
            assert isinstance(self._max_seqs, int) and self._max_seqs > 0

    @staticmethod
    def _parse_batch_size(
        batch_size: Union[int, Dict[str, int], NumbersDict, None, Callable],
        *,
        data_dict: Optional[Dict[str, Any]] = None,
    ) -> Union[NumbersDict, Callable]:
        """
        :param batch_size: see __init__()
        :return: batch_size
        """
        if callable(batch_size):
            if data_dict:
                batch_size = batch_size(**BatchingIterDataPipe._get_user_func_kwargs_from_data_dict(data_dict))
            else:
                return batch_size
        return NumbersDict(sys.maxsize if batch_size is None else batch_size)

    @staticmethod
    def _parse_max_seqs(
        max_seqs: Union[int, None, Callable], *, data_dict: Optional[Dict[str, Any]] = None
    ) -> Union[int, Callable]:
        """
        :param max_seqs: see __init__()
        :return: max_seqs
        """
        if callable(max_seqs):
            if data_dict:
                max_seqs = max_seqs(**BatchingIterDataPipe._get_user_func_kwargs_from_data_dict(data_dict))
            else:
                return max_seqs
        return sys.maxsize if (max_seqs is None or max_seqs == -1) else max_seqs

    @staticmethod
    def _get_user_func_kwargs_from_data_dict(data_dict: Dict[str, Any]) -> Dict[str, Any]:
        epoch = int(data_dict["epoch"])
        seq_idx = int(data_dict["seq_idx"])
        num_seqs = int(data_dict["num_seqs"])  # >=1 if known, otherwise -1
        complete_frac = float(data_dict["complete_frac"])  # >= 0 if known, otherwise -1
        epoch_continuous = (
            epoch - 1 + complete_frac
            if complete_frac >= 0.0
            else (epoch - 1 + (seq_idx + 1) / num_seqs)
            if num_seqs > 0
            else None
        )
        return {"epoch": epoch, "epoch_continuous": epoch_continuous, "seq_idx": seq_idx, **get_fwd_compat_kwargs()}

    def __iter__(self):
        """
        :return: generator providing batches in the form of lists of sequences, where each sequence is a dict
          data_key -> data_array.
        :rtype: Iterable[list[dict[str, numpy.ndarray]]]
        """
        current_batch = []
        current_max_sequence_lengths = NumbersDict(0)  # data_key -> length of longest sequence in current batch

        for data_dict in self._dataset:
            max_seqs = self._parse_max_seqs(self._max_seqs, data_dict=data_dict)
            max_batch_size = self._parse_batch_size(self._max_batch_size, data_dict=data_dict)
            assert isinstance(max_seqs, int) and max_seqs > 0
            assert isinstance(max_batch_size, NumbersDict) and max_batch_size.min_value() > 0

            if len(current_batch) >= max_seqs:
                yield current_batch
                current_batch = []
                current_max_sequence_lengths = NumbersDict(0)

            # TODO: This assumes all data has time as first dimension. Currently we can't know better..
            sequence_lengths = NumbersDict(
                {data_key: data.shape[0] for data_key, data in data_dict.items() if data.shape}
            )

            max_sequence_lengths_if_included = NumbersDict.max([current_max_sequence_lengths, sequence_lengths])
            batch_size_if_included = max_sequence_lengths_if_included * (len(current_batch) + 1)  # including padding

            if current_batch and batch_size_if_included.any_compare(max_batch_size, (lambda a, b: a > b)):
                yield current_batch
                current_batch = [data_dict]
                current_max_sequence_lengths = sequence_lengths
            else:
                current_batch.append(data_dict)
                current_max_sequence_lengths = max_sequence_lengths_if_included

        if current_batch:
            yield current_batch


class BucketOrderingIterDataPipe(torch.utils.data.IterDataPipe):
    """
    Converts a dataset yielding sequences (dict data_key -> array per sequence) into
    a dataset yielding lists of these sequences, i.e. batches. Can be used instead of
    the default ordered batching via the `custom_batching` config key.

    Sequences are grouped into a distinct set of length buckets. Each bucket has a
    limit of how many sequences it can buffer before being sent off to training.

    Note, that batches are not yet merged into a single (padded) data array here,
    this happens in 'collate_batch()'.
    """

    def __init__(
        self,
        dataset: torch.utils.data.IterableDataset,
        *,
        buckets: Sequence[Tuple[int, int]],
        length_key: str,
        random_bucket_prob: float = 0.0,
        seed: Optional[int] = None,
    ):
        """
        :param dataset: dataset to apply bucket batching to
        :param buckets: Bucket configuration as tuples of seq length and max number of seqs in that bucket.
            Segments longer than the largest size limit configured in the buckets are dropped. To avoid dropping
            any segments make sure your largest bucket allows segments larger than your longest training segment.
        :param length_key: data key to take as length measure
        :param random_bucket_prob: Probability of putting a segment not into the best-fitting bucket, but into
            a randomly chosen still-fitting bucket.
            This increases seq length variation within the buckets at the cost of slighly more padding.
        :param seed: random seed
        """
        self._dataset = dataset
        self._length_key = length_key
        assert random_bucket_prob >= 0.0
        self._random_bucket_prob = random_bucket_prob
        self._rng = numpy.random.RandomState()
        self._seed = seed % (2**32) if seed is not None else None

        assert buckets, "empty bucket batching configuration"
        if not all(size > 0 and max_seqs > 0 for size, max_seqs in buckets):
            raise ValueError(f"bucket sizes and max seqs in bucket must be positive")
        self._max_seq_lens, self._max_bucket_sizes = zip(*sorted(buckets))
        assert len(set(self._max_seq_lens)) == len(self._max_seq_lens), "seq len boundaries must all be unique"

    def __iter__(self):
        """:return: generator applying bucket ordering on the data"""
        buckets: List[List[Dict[str, numpy.ndarray]]] = [[] for _ in range(len(self._max_seq_lens))]
        buckets_full_counter = [0 for _ in range(len(buckets))]

        for data_dict in self._dataset:
            data_dict: Dict[str, numpy.ndarray]
            data_size = data_dict[self._length_key].shape[0]
            bucket_idx = bisect.bisect_left(self._max_seq_lens, data_size)
            if bucket_idx >= len(self._max_seq_lens):
                # seg is too long, drop it
                continue
            if (
                self._random_bucket_prob > 0.0
                and bucket_idx < len(self._max_seq_lens) - 1
                and self._rng.rand() < self._random_bucket_prob
            ):
                bucket_idx = self._rng.randint(bucket_idx, len(self._max_bucket_sizes))
            buckets[bucket_idx].append(data_dict)
            if len(buckets[bucket_idx]) >= self._max_bucket_sizes[bucket_idx]:
                yield buckets[bucket_idx]
                buckets[bucket_idx] = []
                buckets_full_counter[bucket_idx] += 1

        non_empty_buckets = [b for b in buckets if b]
        yield from non_empty_buckets

        # we do not count the buckets w/ leftover data as they were not completely filled
        description_str = ", ".join(f"{limit}: {cnt}" for limit, cnt in zip(self._max_seq_lens, buckets_full_counter))
        print(f"Batching buckets full: {description_str}", file=log.v4)

    def __getitem__(self, index):
        raise Exception(f"{self.__class__.__name__}.__getitem__ is not supported")

    def set_seed(self, seed: int) -> BucketOrderingIterDataPipe:
        """
        Sets the seed for the next invocation of ``__iter__``, for compatibility with
        ``torch.utils.data.graph_settings.apply_random_seed``.
        """
        self._seed = seed % (2**32)  # seed must be within [0, 2**32) for seeding RandomState
        return self

    def reset(self):
        """resets the internal state of the data pipe"""
        if self._seed is None:
            self._seed = int(2**31 + torch.empty((), dtype=torch.int32).random_().item())
        self._rng.seed(self._seed)
        self._seed = None


def get_batching_iterable_dataset_from_config(
    *, dataset: torch.utils.data.IterableDataset, config: Config, train: bool
) -> torch.utils.data.IterableDataset:
    """
    Batches the segments in the given dataset given the config settings.

    :param dataset: dataset whose segments to group into batches
    :param config: RETURNN config
    :param train: whether we are in training or in inference mode
    """
    torch_batching = config.typed_value("torch_batching", None)
    if torch_batching is None:
        batch_size = config.typed_value("batch_size", -1)
        batch_size = config.typed_value(f"batch_size_{'train' if train else 'dev'}", batch_size)
        assert batch_size != -1, f"batch_size or batch_size_{'train' if train else 'dev'} not defined in config"
        max_seqs = config.typed_value("max_seqs", -1)
        batches_dataset = BatchingIterDataPipe(dataset, batch_size=batch_size, max_seqs=max_seqs)
        return batches_dataset

    if isinstance(torch_batching, dict):
        assert "class" in torch_batching
        batching_args = torch_batching.copy()
        type_name = batching_args.pop("class")
        assert isinstance(type_name, str)
        cls = globals()[type_name]
    elif callable(torch_batching):
        # callables need to be forward compatible
        batching_args = get_fwd_compat_kwargs()
        batching_args["train"] = train
        cls = torch_batching
    else:
        raise ValueError(
            f"custom_batching must either be a dict containing a `class` key naming a type, a type or a callable."
        )
    batches_dataset = cls(dataset, **batching_args)
    assert isinstance(batches_dataset, torch.utils.data.IterableDataset)
    return batches_dataset


class LenFilterDataPipe(torch.utils.data.IterDataPipe):
    """
    Removes sequences which are either too long or too short from a dataset
    Returns dataset yielding list of data lengths within the defined range
    """

    def __init__(
        self,
        dataset: torch.utils.data.IterableDataset,
        min_seq_length: Union[int, NumbersDict] = None,
        max_seq_length: Union[int, NumbersDict] = None,
    ):
        """
        :param dataset: dataset to apply the filter to
        :param min_seq_length: minimum sequence length either in general or per data_key via dict
        :param max_seq_length: maximum sequence length either in general or per data_key via dict
        """
        super().__init__()
        self._dataset = dataset
        self._min_seq_length = NumbersDict(0 if min_seq_length is None else min_seq_length)
        self._max_seq_length = NumbersDict(sys.maxsize if max_seq_length is None else max_seq_length)

    def __iter__(self):
        """
        :return: generator providing filtered data where each sequence is a dict
          data_key -> data_array.
        :rtype: Iterable[dict[str, numpy.ndarray]]
        """
        for data_dict in self._dataset:
            # TODO: This assumes all data has time as first dimension. Currently we can't know better..
            sequence_lengths = NumbersDict(
                {data_key: data.shape[0] for data_key, data in data_dict.items() if data.shape}
            )
            if sequence_lengths.any_compare(self._min_seq_length, lambda a, b: a < b):
                continue
            if sequence_lengths.any_compare(self._max_seq_length, lambda a, b: a > b):
                continue
            yield data_dict

    def __getitem__(self, index):
        raise Exception(f"{self.__class__.__name__}.__getitem__ not supported")


class ShufflingDataPipe(torch.utils.data.IterDataPipe):
    """
    Data pipe that is similar to ``torch.utils.data.datapipes.iter.Shuffler``,
    but it will keep certain data keys of the batches in order while shuffling the rest.

    Used for e.g. ``complete_frac`` and ``seq_idx``.
    """

    def __init__(
        self,
        dataset: torch.utils.data.IterableDataset,
        *,
        buffer_size: int,
        monotonic_data_keys: Sequence[str],
        seed: Optional[int] = None,
    ):
        """
        :param dataset: batches dataset to shuffle
        :param buffer_size: buffer size for shuffling
        :param monotonic_data_keys: data keys that will be excluded from shuffling/keep their order
        :param seed: random seed
        """
        super().__init__()

        self._dataset = dataset
        self._buffer: List[List[Dict[str, Any]]] = []
        self._next_buffer: List[List[Dict[str, Any]]] = []
        assert buffer_size > 0
        self._buffer_size = buffer_size
        self._monotonic_data_keys = monotonic_data_keys
        self._rng = numpy.random.RandomState()
        self._seed = seed % (2**32) if seed is not None else None

    def __iter__(self):
        # The implementation is very similar to the PostprocessingDataset's combinator LaplaceOrdering.

        data_iter = iter(self._dataset)

        self._buffer.extend(itertools.islice(data_iter, self._buffer_size))
        has_ended = False

        while True:
            # Make sure to not reorder the monotonic values from self._monotonic_data_keys.
            # These can contain things like complete_frac, which should be kept in order.
            ordered_data = {
                key: [data_dict[key] for batch in self._buffer for data_dict in batch]
                for key in self._monotonic_data_keys
            }
            self._rng.shuffle(self._buffer)
            for key in self._monotonic_data_keys:
                data_dicts = [data_dict for batch in self._buffer for data_dict in batch]
                assert len(data_dicts) == len(ordered_data[key])
                for ordered_value, data_dict in zip(ordered_data[key], data_dicts):
                    data_dict[key] = ordered_value

            for item in self._buffer:
                yield item

                try:
                    if not has_ended:
                        self._next_buffer.append(next(data_iter))
                except StopIteration:
                    has_ended = True

            if len(self._buffer) < self._buffer_size:
                assert has_ended and not self._next_buffer
                break

            self._buffer.clear()
            self._buffer, self._next_buffer = self._next_buffer, self._buffer

    def set_seed(self, seed: int) -> ShufflingDataPipe:
        """
        Sets the seed for the next invocation of ``__iter__``, for compatibility with
        ``torch.utils.data.graph_settings.apply_random_seed``.
        """
        self._seed = seed % (2**32)  # seed must be within [0, 2**32) for seeding RandomState
        return self

    def reset(self):
        """resets the internal state of the data pipe"""
        self._buffer.clear()
        self._next_buffer.clear()
        if self._seed is None:
            self._seed = int(2**31 + torch.empty((), dtype=torch.int32).random_().item())
        self._rng.seed(self._seed)
        self._seed = None

    def __getstate__(self):
        state = (
            self._dataset,
            self._buffer,
            self._next_buffer,
            self._buffer_size,
            self._monotonic_data_keys,
            self._rng.get_state(),
            self._seed,
        )
        if torch.utils.data.IterDataPipe.getstate_hook is not None:
            return torch.utils.data.IterDataPipe.getstate_hook(state)
        return state

    def __setstate__(self, state):
        (
            self._dataset,
            self._buffer,
            self._next_buffer,
            self._buffer_size,
            self._monotonic_data_keys,
            rng_state,
            self._seed,
        ) = state
        self._rng = numpy.random.RandomState()
        self._rng.set_state(rng_state)

    def __getitem__(self, index):
        raise Exception(f"{self.__class__.__name__}.__getitem__ not supported")


def create_data_loader_from_batches(
    batches_dataset: torch.utils.data.Dataset, loader_opts: Optional[Dict[str, Any]] = None
) -> torch.utils.data.DataLoader:
    """
    Create DataLoader based on dataset over batches, e.g. via :class:`BatchingIterDataPipe`.
    """
    if loader_opts is None:
        loader_opts: Dict[str, Any] = {}
    else:
        loader_opts = loader_opts.copy()

    # Make sure to use workers unless specified otherwise to ensure reasonable GPU
    # utilization and to work around some issues wrt. overflowing ulimits when
    # workers are non-persistent.
    #
    # See for context:
    # - https://github.com/rwth-i6/returnn/issues/1560
    # - https://github.com/pytorch/pytorch/issues/129868
    loader_opts.setdefault("num_workers", 1)

    if loader_opts.get("num_workers"):
        loader_opts.setdefault("persistent_workers", True)
        loader_opts["worker_init_fn"] = _DataLoaderWorkerInitFunc(
            other_worker_init_fn=loader_opts.get("worker_init_fn")
        )
        # We don't want to use the default fork start method
        # (https://github.com/rwth-i6/returnn/issues/1494 and potentially lots of other issues with fork).
        # We cannot use the standard spawn start method, as DataLoader would start daemonic processes,
        # which is incompatible with some of our datasets
        # (https://github.com/rwth-i6/returnn/issues/1495).
        loader_opts.setdefault("multiprocessing_context", "spawn_non_daemonic")
        if loader_opts.get("multiprocessing_context") == "spawn_non_daemonic":
            from returnn.config import SubProcCopyGlobalConfigPreInitFunc
            from returnn.util.multi_proc_non_daemonic_spawn import NonDaemonicSpawnContext

            loader_opts["multiprocessing_context"] = NonDaemonicSpawnContext(
                # It's important that this init function is called even before the unpickling of the dataset,
                # as that unpickling might depend on the right context, e.g. having a global RETURNN config.
                # See _DataLoaderWorkerPreInitFunc below, https://github.com/rwth-i6/returnn/issues/1495.
                process_pre_init_func=SubProcCopyGlobalConfigPreInitFunc()
            )

    return torch.utils.data.DataLoader(
        batches_dataset,
        collate_fn=collate_batch,
        # Batching is already done by BatchingIterDataPipe.
        batch_size=None,
        # Explicitly not use the following opts, which are not supported and/or do not make sense
        # for an iterable-style dataset.
        shuffle=None,
        sampler=None,
        batch_sampler=None,
        # User-defined
        **loader_opts,
    )


class _DataLoaderWorkerInitFunc:
    def __init__(self, *, other_worker_init_fn: Optional[Callable] = None):
        self.other_worker_init_fn = other_worker_init_fn

    def __call__(self, worker_id: int):
        if sys.platform == "linux":
            with open("/proc/self/comm", "w") as f:
                f.write(f"TDL worker {worker_id}")

        if self.other_worker_init_fn:
            self.other_worker_init_fn(worker_id)
